/* $Id: clientcomm.hpp,v 1.18 2004/09/06 21:05:21 chfreund Exp $ */

#ifndef _CLIENTCOMM_HPP_
#define _CLIENTCOMM_HPP_

#include <map>
#include <set>
#include <SDL.h>
#include <SDL_net.h>
#include "comm.hpp"
#include "string.hpp"

class Player;
class World;
class Client;

//! Describes a server by its name and address
struct ServerEntry {
	//! Name of the server
	char name[256];
	//! Address of the server
	IPaddress address;
	//! Theme running on the server
	String theme;
	//! Number of players on the server
	Uint8 numberPlayers;
	
	//! Constructor
	ServerEntry() {
		name[0] = 0;
		address.host = 0;
		address.port = 0;
	}
	
	//! Constructor
	ServerEntry( const char* serverName ) {
		strncpy( name, serverName, 256 );
		address.host = 0;
		address.port = 0;
	}
}; // ServerEntry

namespace std {
template <>
struct less<ServerEntry> {
	bool operator()( const ServerEntry& e1, const ServerEntry& e2 ) const {
		return ( e1.address.host < e2.address.host ||
		         ( e1.address.host == e2.address.host &&
		           e1.address.port < e2.address.port ) );
	}
};
}

//////////////////////////////////////////////////////////////////////
// A specialized communicator class for the client
//////////////////////////////////////////////////////////////////////

/*! \brief A specialized communicator class for the client
 */

class ClientComm : public Comm {
	private:
		//! The TCP socket for the communication with the server
		TCPsocket m_tcpSocket;
		//! A socket set for testing the TCP socket for activity
		SDLNet_SocketSet m_tcpSocketSet;
		//! The UDP socket for the communication with the server
		UDPsocket m_udpSocket;
		//! The port number of the UDP socket
		Uint16 m_localPort;
		
		//! The server address
		ServerEntry m_server;
		
		//! The frame number of the last consecutive echo message received 
		Uint8 m_lastReceivedFrame;
		//! A buffer for echo messages
		std::map<Uint8, EchoData*> m_echoBuffer;
		
		//! The client that is using this communicator
		Client* m_client;
		
		//! flag for connection status
		bool m_isConnected;

	public:
		//! Constructor
		ClientComm( Client* client );
		//! Destructor
		~ClientComm();
		
		/*! \brief Look for game servers in the local network
		 *  \return A vector with all hosts in the local network where
		 *    a game server is running
		 */
		std::set<ServerEntry> lookForServer();
		/*! \brief Sets the server but does not connect
		 *  \param entry The server to be set
		 *  \return \b true on success, \b false on failure
		 */
		bool setServer( ServerEntry& entry );
		/*! \brief Establishes a connection
		 *
		 *  This method connects a player with the server that has been
		 *  set before and receives the world
		 *  \param player The player that should be connected
		 *  \param world The world object that will be filled in on
		 *    successful connection
		 *  \return \b true on success, \b false on failure
		 */
		bool connect( Player* player, World*& world );
		/*! \brief Closes current connection
		 *
		 * Additionally performs cleanup regarding the old connection
		 */
		void disconnect();
		/*! \brief Sends a single event to the server
		 *  \param event The event to be sent
		 */
		void sendEvent( const Event& event );
		/*! \brief Get the echo data for a given frame
		 *  \param frame The frame number of the expected echo message
		 *  \return The echo data on success, 0 if the wanted echo message
		 *    has not been received yet (not necessarily a failure)
		 *  \note The returned EchoData object has been created dynamically
		 *    so it must be freed sometime
		 */
		EchoData* getEcho( Uint8 frame );
		/*! Checks the TCP socket for activity (which indicates
		 *  that another player has joined the game or a chat
		 *  message has arrived) and perform the necessary actions
		 */
		void handleTCPMessages();
		/*! \brief Send a chat message to the server
		 *  \param message The chat message to be broadcasted
		 */
		void sendChatMessage( String& message, const SDL_Color& color );
		
		bool isConnected() const { return m_isConnected; }
		
	private:
		/*! \brief Receives an echo from the server
		 *  \param echo The echo message will be stored in \em echo
		 *  \return \b true on success, \b false on failure
		 */
		bool receiveEcho( EchoData& echo );

}; // ClientComm

#endif // _CLIENTCOMM_HPP_
