/* $Id: collidableobject.hpp,v 1.19 2005/04/11 16:14:27 chfreund Exp $ */

#ifndef _COLLIDABLEOBJECT_HPP_
#define _COLLIDABLEOBJECT_HPP_

#include "attachableobject.hpp"
#include "particles.hpp"

/**********************************************************/

class Player;

/**********************************************************/

/*!
Collidable objects are derived from the Object class. Their additional
feature is that they can collide with other collidable objects, earth and/or
particles depending on the implementation of CollidableObject::update.
*/

class CollidableObject : public AttachableObject {
	public:
		CollidableObject( void );
		virtual ~CollidableObject( void );

	//////////
	// Health & Damage

		/*! This method is called to determine the collision with another
		    collidable object or a particle at position \a x, \a c
		    \return \c true if the specified position is on or within the collision
	              rectangle of the object; \c false otherwise.
		*/
		bool testHit( const int x, const int y ) {
			if( ! m_collRectSet ) return false;

			const bool hit =
			        x >= ROUND( m_positionX ) + m_collRectX &&
			        x <= ROUND( m_positionX ) + m_collRectX + m_collRectWidth - 1 &&
			        y >= ROUND( m_positionY ) + m_collRectY &&
			        y <= ROUND( m_positionY ) + m_collRectY + m_collRectHeight - 1;

			DBG( 5 ) {
				if( ! hit ) {
					const real dx = m_positionX - x,
					            dy = m_positionY - y;
					INFO( "CollidableObject::testHit:no hit, dist = %6.1f\n"
					      "\tx : %3i, y : %3i\n"
					      "\tx1: %3i, y1: %3i\n"
					      "\tx2: %3i, y2: %3i\n",
					      SQRT_REAL( dx*dx + dy*dy ), x, y,
					      ROUND( m_positionX ) + m_collRectX,
					      ROUND( m_positionY ) + m_collRectY,
					      ROUND( m_positionX ) + m_collRectX + m_collRectWidth  - 1,
					      ROUND( m_positionY ) + m_collRectY + m_collRectHeight - 1 );
				}
			}

			return hit;
		}

		bool testGrounded( void );

		/*! This method is called when the object is damaged by the specified
		    particle \a p.
				Things that have to be done within this method:
				- add recoil to the object's speed
				- calculate damage and subtract this value from \a m_health
				- play damage sample and set \a m_damageSampleTimeout to an appropriate value
				- destroy object if necessary
			\return the affected damage

		    \todo Move the basic stuff into the base class CollidableObject.
		*/
		virtual int applyDamage( const Particles::ParticleData& p );
		virtual int applyDamage( const int damage, Player* shooter );

		Sint32 getHealth( void ) const { return m_health; }
		real getMass( void ) const { return m_mass; }
		Sint32 getCollRectX( void ) const { return m_collRectX; }
		Sint32 getCollRectY( void ) const { return m_collRectY; }
		Sint32 getCollRectWidth( void ) const { return m_collRectWidth; }
		Sint32 getCollRectHeight( void ) const { return m_collRectHeight; }

		void setHealth( const Sint32 health ) { m_health = health; }
		void setCollRect( void );
		void removeCollRect( void );

		virtual void serialize( Uint8*& bufferPointer ) const;
		virtual void deserialize( Uint8*& bufferPointer );
		virtual Uint32 getSerializeBufferSize() const;

	protected:
		real  m_mass;
		Sint32 m_collRectX, m_collRectY,
		       m_collRectWidth, m_collRectHeight;
		Sint32 m_health;
		Sint32 m_damageSampleTimeout;
		bool m_collRectSet;
		bool m_firstUpdate;
};

/**********************************************************/

#endif // _COLLIDABLEOBJECT_HPP_
