/* $Id: comm.hpp,v 1.26 2005/03/16 09:53:51 chfreund Exp $ */

#ifndef _COMM_HPP_
#define _COMM_HPP_

#include <SDL_types.h>
#include <SDL_net.h>
#include "packet.hpp"
#include "udpdata.hpp"

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
#define ENCODE_IP(a,b,c,d) (((((((d) << 8) + (c)) << 8) + (b)) << 8) + (a))
#define DECODE_IP(h) (255 & h), (255 & ((h) >> 8)), \
                 (255 & ((h) >> 16)), (255 & ((h) >> 24))
#else // SDL_BYTEORDER == SDL_BIG_ENDIAN
#define ENCODE_IP(a,b,c,d) (((((((a) << 8) + (b)) << 8) + (c)) << 8) + (d))
#define DECODE_IP(h) (255 & ((h) >> 24)), (255 & ((h) >> 16)), \
                     (255 & ((h) >> 8)), (255 & h)
#endif


//////////////////////////////////////////////////////////////////////
// Base class for communicators
//////////////////////////////////////////////////////////////////////

/*! \brief Base class for communicators
 *
 *  Among the common data members used in the derived classes this
 *  class provides the functions tcpSendPacket() and tcpRecvPacket()
 *  for easy communication
 */

class Comm {
	protected:
		//! UDP (TCP) port number for incoming bait packets (client
		//! connections)
		static const Uint16 m_SERVER_PORT = 0xb8b8;
		//! The maximum number of frames a client is allowed to be
		//! lagged
		static const int m_MAX_SERVER_LAG = 100;
		//! If a client sends more than m_MAX_CLIENT_LAG packets where the
		//! number of the last received frame of the client is the same then
		//! the server assumes that the client missed the following echo packet
		//! and resends it.
		static const int m_MAX_CLIENT_LAG = 1;
		//! The size of the buffer for echo packets
		static const int N_ECHO_BUFFERS = 256;
		
		//! The local address of the host
		IPaddress m_localAddress;
		//! UDP packet for discovering the server
		UDPpacket* m_baitPacket;
		//! UDP packet with event information
		UDPpacket* m_eventPacket;
		//! UDP answer packet from the server
		UDPpacket* m_echoPacket;
		
		//! UDP data for a client event
		EventData m_eventData;
		//! UDP data for an echo message
		EchoData m_echoData;
		
	public:
		//! Constructor
		Comm ();
		//! Destructor
		~Comm ();
		
		//! \return pointer to the UDP data with the current echo message
		const EchoData* getEchoData() const {
			return &m_echoData;
		}
		
		/*! \brief Send a TCP packet over the given socket
		 *  \param socket The TCP socket over which the data is to be sent
		 *  \param packet Pointer to the TCP packet to be sent
		 *  \return 0 on success, -1 on failure
		 */ 
		int tcpSendPacket( TCPsocket socket, TCPPacket* packet );
		/*! \brief Receive a TCP packet over the given socket
		 *  \param socket The TCP socket for the reception of the packet
		 *  \return 0 on failure else the received packet
		 *  \note The received packet is allocated dynamically so it has to be
		 *    deleted when it is no longer used.
		 */
		TCPPacket* tcpRecvPacket( TCPsocket socket );
		
		/*! \brief Send a serializable object over the given socket
		 *  \param socket The TCP socket over which the object is to be sent
		 *  \param object The object to be sent
		 *  \note On the receiving side the function tcpRecvObject() with an
		 *    object of the same class as \em object has to be called in order
		 *    to work
		 */
		void tcpSendObject( TCPsocket socket, Serializable* object );
		/*! \brief Receive a serializable object from the given socket
		 *  \param socket The TCP socket for the reception of the object
		 *  \param object A pointer to an object of the same class as the
		 *    object that is expected to be received
		 *  \note If an object of another class than expected is received
		 *    everything will be messed up so it is better to avoid it ;-)
		 */
		void tcpRecvObject( TCPsocket socket, Serializable* object );
};

#endif // _COMM_HPP_
