/* $Id: grenade.cpp,v 1.27 2005/04/08 13:50:42 pohlt Exp $ */

#include "grenade.hpp"
#include "world.hpp"
#include "particles.hpp"
#include "random.hpp"

/**********************************************************/

Mix_Chunk* Grenade::m_explodeSample, * Grenade::m_bounceSample,
         * Grenade::m_damageSample;

/**********************************************************/

Grenade::Grenade() {
	m_health = 180000;
	m_mass = 50.0;
	m_collRectX = -2; m_collRectY = -3;
	m_collRectWidth = 6; m_collRectHeight = 6;
}

/**********************************************************/

bool Grenade::initialize( const Vector& pos, const Vector& vel,
                          const Uint8 owner, const Sint32 ttl,
                          const Sint32 nShrapnels, const Sint32 damage ) {
	int dummy;

	if( ! m_worldPointer->getMap()->testFilledCollRect( ROUND( pos.x ) + m_collRectX,
	                                                    ROUND( pos.y ) + m_collRectY,
	                        m_collRectWidth, m_collRectHeight, dummy, dummy ) ) {
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY GRENADE DATA MEMBERS AFTER THIS POINT
		return false;
	}

	m_positionX = pos.x;
	m_positionY = pos.y;

	m_speedX = vel.x;
	m_speedY = vel.y;

	m_owner = owner;
	m_ttl = ttl;
	m_nShrapnels = nShrapnels;
	m_damage = damage;

	// set grenade stencil
	setCollRect();

	return true;
}

/**********************************************************/

void Grenade::update() {
	if( m_damageSampleTimeout > 0 ) m_damageSampleTimeout--;

	Map* const map = m_worldPointer->getMap();

	const real DT = 1.0;

	int cx, cy, pcx, pcy;

	// remove grenade stencil
	removeCollRect();

	if( isGrounded() && ( ! testGrounded() || getSpeed().abs2() > 0.01) ) {
		setState( GROUNDED, false );
	}

	if( isGrounded() ) {
		m_speedX = m_speedY = 0;
	}
	else {
		// move grenade
		m_speedY += GRAVITATION*DT;

		real newPosX = m_positionX+m_speedX*DT,
		      newPosY = m_positionY+m_speedY*DT;

		// is there something to collide?
		if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
		                           ROUND( m_positionY ) + m_collRectY,
		                           ROUND( newPosX     ) + m_collRectX,
		                           ROUND( newPosY     ) + m_collRectY,
		                           m_collRectWidth, m_collRectHeight,
		                           pcx, pcy, cx, cy ) ) {
			// no, then just keep on flying
			m_positionX = newPosX;
			m_positionY = newPosY;
		}
		else {
			// play bounce sound
			m_positionX = pcx - m_collRectX;
			m_positionY = pcy - m_collRectY;
			if( ! m_firstUpdate ) {
				Audio::getInstance()->playSound( m_bounceSample, ROUND( m_positionX ),
																												 ROUND( m_positionY ) );
				const real speed2 = m_speedX*m_speedX + m_speedY*m_speedY;

				if( speed2 < 5.0 && testGrounded() ) {
					setState( GROUNDED, true );
					m_speedX = m_speedY = 0;
				}
				else {
					const real speed = SQRT_REAL( speed2 ),
					            outAngle = map->calculateBounceAngle( m_speedX, m_speedY, cx, cy, Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT );
					m_speedX = speed * (speed > 5.0 ? 0.5 : 0.9) * COS_REAL( outAngle );
					m_speedY = speed * (speed > 5.0 ? 0.5 : 0.9) * SIN_REAL( outAngle );
				}
			}
		}
	}

	// explode?
	if( m_health <= 0 || (m_ttl >= 0 && --m_ttl <= 0) ) {
		Audio::getInstance()->playSound( m_explodeSample, ROUND( m_positionX ),
		                                                  ROUND( m_positionY ) );
		m_worldPointer->newObject( EXPLOSION_60 )->setPosition( ROUND( m_positionX ),
		                                                        ROUND( m_positionY ));
		for( int s = 0; s < m_nShrapnels; s++ ) {
			m_worldPointer->getParticles()->addParticle( Particles::ParticleData( getPosition(),
			     globalRnd.getVector( 25 + 5*globalRnd.getNormedReal() ) + getSpeed(),
			     Particles::EXPL_EARTH | Particles::EXPL_OBJECT | Particles::DAMAGE | Particles::EXPL_GEN_SPARKS,
			     0, m_owner, 5, m_damage, 0x009900 ));
		}

		// remove grenade object
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY GRENADE DATA MEMBERS AFTER THIS POINT
		return;
	}

	// set grenade stencil
	setCollRect();

	m_firstUpdate = false;

	DBG( 5 ) map->setColorAt( ROUND( m_positionX ),
	                          ROUND( m_positionY ), 0xffffff );
}

/**********************************************************/

int Grenade::applyDamage( const Particles::ParticleData& p ) {
	const int damage = CollidableObject::applyDamage( p );

	// create sparks
	for( int i = 0; i < 2; i++ ) {
		m_worldPointer->getParticles()->addParticle( Particles::ParticleData( p.pos,
		     globalRnd.getVector( 1 + 7*globalRnd.getNormedReal() ) + getSpeed(),
		     Particles::BOUNCE_EARTH | Particles::BOUNCE_OBJECT | Particles::REDUNDANT | Particles::SPARK, 0, 0,
		     Particles::m_N_SPARK_COLORS/2 + (globalRnd.getUint32() % (Particles::m_N_SPARK_COLORS/2+1)),
		     0, 0xffffff));
	}

	if( damage > 50 && m_damageSampleTimeout == 0 ) {
		Audio::getInstance()->playSound( m_damageSample, ROUND( m_positionX ),
		                                                 ROUND( m_positionY ) );
		m_damageSampleTimeout = 10;
	}

	return damage;
}

/**********************************************************/

void Grenade::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );

	CollidableObject::serialize( bufferPointer );	

	Serialize<Uint8 >::serialize( m_owner,      bufferPointer );
	Serialize<Sint32>::serialize( m_ttl,        bufferPointer );
	Serialize<Sint32>::serialize( m_nShrapnels, bufferPointer );
	Serialize<Sint32>::serialize( m_damage,     bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Grenade );
}

/**********************************************************/

void Grenade::deserialize( Uint8*& bufferPointer ) {
	CollidableObject::deserialize( bufferPointer );

	Serialize<Uint8 >::deserialize( bufferPointer, m_owner );
	Serialize<Sint32>::deserialize( bufferPointer, m_ttl );
	Serialize<Sint32>::deserialize( bufferPointer, m_nShrapnels );
	Serialize<Sint32>::deserialize( bufferPointer, m_damage );

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

Uint32 Grenade::getSerializeBufferSize() const {
	return   CollidableObject::getSerializeBufferSize()
	       + Serialize<Uint8>::sizeOf( m_owner )
	       + Serialize<Sint32>::sizeOf( m_ttl )
	       + Serialize<Sint32>::sizeOf( m_nShrapnels )
	       + Serialize<Sint32>::sizeOf( m_damage )
	         // adds additional size for debugging (see serialize.hpp),
	         // but only, if the tags are used
	         PLUS_TAG_SIZE( 1 );
}

/**********************************************************/
