/* $Id: missile.cpp,v 1.20 2005/04/08 13:16:03 pohlt Exp $ */

#include "missile.hpp"
#include "world.hpp"
#include "particles.hpp"
#include "random.hpp"

/**********************************************************/

Mix_Chunk* Missile::m_explodeSample,
         * Missile::m_damageSample;

/**********************************************************/

Missile::Missile() {
	m_health = 300000;
	m_mass = 100.0;
	m_collRectX = -4; m_collRectY = -4;
	m_collRectWidth = 9; m_collRectHeight = 9;
}

/**********************************************************/

bool Missile::initialize( const Vector& pos, const Vector& vel,
                          const Uint8 owner,
                          const Sint32 nShrapnels, const Sint32 damage,
                          const Sint32 accelTimer, const Vector& thrust ) {
	int dummy;

	if( ! m_worldPointer->getMap()->testFilledCollRect( ROUND( pos.x ) + m_collRectX,
	                                                    ROUND( pos.y ) + m_collRectY,
	                        m_collRectWidth, m_collRectHeight, dummy, dummy ) ) {
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY MISSILE DATA MEMBERS AFTER THIS POINT
		return false;
	}

	m_positionX = pos.x;
	m_positionY = pos.y;

	m_owner = owner;
	m_nShrapnels = nShrapnels;
	m_damage = damage;
	m_accelTimer = accelTimer;
	m_thrust = thrust;

	m_speedX = thrust.x;
	m_speedY = thrust.y;
	updateSpriteSelection();
	m_speedX = vel.x;
	m_speedY = vel.y;

	// set missile stencil
	setCollRect();

	return true;
}

/**********************************************************/

void Missile::update() {
	if( m_damageSampleTimeout > 0 ) m_damageSampleTimeout--;

	Map* const map = m_worldPointer->getMap();

	const real DT = 1.0;

	int cx, cy, pcx, pcy;

	// remove missile stencil
	removeCollRect();

	// accelerate missile
	if( m_accelTimer != 0 ) {
		m_speedX += m_thrust.x;
		m_speedY += m_thrust.y;
		// throw out some booster sparks
		for( int i = 0; i < 15; i++ ) {
			m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
			     getPosition(),
			     getSpeed() - m_thrust*10.0 + globalRnd.getVector( 1.5 ),
			     Particles::BOUNCE_EARTH | //Particles::BOUNCE_OBJECT |
			     Particles::REDUNDANT | Particles::SPARK, 2, 0,
			     Particles::m_N_SPARK_COLORS/2 + globalRnd.getUint32() % (Particles::m_N_SPARK_COLORS/2+1), 1,
			     0xffffff));
		}
		// generate a smoke cloud
		m_worldPointer->newObject( SMOKE_30 )->
			setPosition( ROUND(m_positionX) - 3 + (globalRnd.getUint32()%7),
		                 ROUND(m_positionY) - 3 + (globalRnd.getUint32()%7) );
		// decrement "fuel"
		if( m_accelTimer > 0 ) m_accelTimer--;
	}
	// move missile
	m_speedY += GRAVITATION*DT;

	real newPosX = m_positionX+m_speedX*DT,
	      newPosY = m_positionY+m_speedY*DT;

	// is there something to collide?
	if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
	                           ROUND( m_positionY ) + m_collRectY,
	                           ROUND( newPosX     ) + m_collRectX,
	                           ROUND( newPosY     ) + m_collRectY,
	                           m_collRectWidth, m_collRectHeight,
	                           pcx, pcy, cx, cy ) ) {
		// no, then just keep on flying
		m_positionX = newPosX;
		m_positionY = newPosY;
	}
	else {
		m_positionX = pcx - m_collRectX;
		m_positionY = pcy - m_collRectY;
		if( ! m_firstUpdate ) m_health = 0; // explode if not first update
	}

	// explode?
	if( m_health <= 0 ) {
		Audio::getInstance()->playSound( m_explodeSample, ROUND( m_positionX ),
		                                                  ROUND( m_positionY ) );
		m_worldPointer->newObject( EXPLOSION_80 )->setPosition( ROUND( m_positionX ),
		                                                        ROUND( m_positionY ));
		for( int s = 0; s < m_nShrapnels; s++ ) {
			m_worldPointer->getParticles()->addParticle( Particles::ParticleData( getPosition(),
			     globalRnd.getVector( 15 + 5*globalRnd.getNormedReal() ), // + getSpeed(),  otherwise shrapnels will just fly forward
			     Particles::EXPL_EARTH | Particles::EXPL_OBJECT | Particles::DAMAGE |
			     Particles::INVISIBLE,
			     0, m_owner, 2, m_damage, 0xffffff ));
		}

		// remove missile object
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY MISSILE DATA MEMBERS AFTER THIS POINT
		return;
	}

	// increment the animation frame
	if( ++m_Frame >= m_SequenceLength ) m_Frame = 0;
	// select new missile sprite
	if( m_accelTimer == 0 ) updateSpriteSelection();

	// set missile stencil
	setCollRect();

	m_firstUpdate = false;

	DBG( 5 ) map->setColorAt( ROUND( m_positionX ),
	                          ROUND( m_positionY ), 0xffffff );
}

/**********************************************************/

int Missile::applyDamage( const Particles::ParticleData& p ) {
	const int damage = CollidableObject::applyDamage( p );

	// create sparks
	for( int i = 0; i < 2; i++ ) {
		m_worldPointer->getParticles()->addParticle( Particles::ParticleData( p.pos,
		     globalRnd.getVector( 1 + 7*globalRnd.getNormedReal() ) + getSpeed(),
		     Particles::BOUNCE_EARTH | Particles::BOUNCE_OBJECT | Particles::REDUNDANT | Particles::SPARK, 0, 0,
		     Particles::m_N_SPARK_COLORS/2 + (globalRnd.getUint32() % (Particles::m_N_SPARK_COLORS/2+1)),
		     0, 0xffffff));
	}

	if( damage > 50 && m_damageSampleTimeout == 0 ) {
		Audio::getInstance()->playSound( m_damageSample, ROUND( m_positionX ),
		                                                 ROUND( m_positionY ) );
		m_damageSampleTimeout = 10;
	}

	if( m_health > 0 ) {
		updateSpriteSelection();
	}

	return damage;
}

/**********************************************************/

void Missile::updateSpriteSelection() {
	const real angle = ATAN2_REAL( m_speedY, m_speedX );
	// set the sequence index.
	// NOTE: setSequence also updates m_SequenceLength.
	setSequence( (ROUND( angle*(18.0/M_PI) - 0.5 )+36) % 36 );

	// only reset the frame index, if we cannot continue
	// this movement in a different aiming angle
	if( getFrame() >= m_SequenceLength )  setFrame( 0 );
}

/**********************************************************/

void Missile::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	CollidableObject::serialize( bufferPointer );	

	Serialize<Uint8 >::serialize( m_owner,      bufferPointer );
	Serialize<Sint32>::serialize( m_nShrapnels, bufferPointer );
	Serialize<Sint32>::serialize( m_damage,     bufferPointer );
	Serialize<Sint32>::serialize( m_accelTimer, bufferPointer );
	Serialize<Vector>::serialize( m_thrust,     bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Missile );
}

/**********************************************************/

void Missile::deserialize( Uint8*& bufferPointer ) {
	CollidableObject::deserialize( bufferPointer );

	Serialize<Uint8 >::deserialize( bufferPointer, m_owner );
	Serialize<Sint32>::deserialize( bufferPointer, m_nShrapnels );
	Serialize<Sint32>::deserialize( bufferPointer, m_damage );
	Serialize<Sint32>::deserialize( bufferPointer, m_accelTimer );
	Serialize<Vector>::deserialize( bufferPointer, m_thrust );

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

Uint32 Missile::getSerializeBufferSize() const {
	return   CollidableObject::getSerializeBufferSize()
	       + Serialize<Uint8>::sizeOf( m_owner )
	       + Serialize<Sint32>::sizeOf( m_nShrapnels )
		   + Serialize<Sint32>::sizeOf( m_damage )
	       + Serialize<Sint32>::sizeOf( m_accelTimer )
		   + Serialize<Vector>::sizeOf( m_thrust )
	         // adds additional size for debugging (see serialize.hpp),
	         // but only, if the tags are used
	         PLUS_TAG_SIZE( 1 );
}

/**********************************************************/
