/* $Id: packet.hpp,v 1.19 2004/10/05 11:05:36 chfreund Exp $ */

#ifndef _PACKET_HPP_
#define _PACKET_HPP_

#include <SDL_types.h>

#include "serializable.hpp"
#include "serialize.hpp"
#include "player.hpp"
#include "world.hpp"
#include "string.hpp"

/*! \brief Base class for packets that are sent from the client to the server
 *    over a TCP connection
 *
 *  What to do if you want to create a new packet class:
 *    -# Derive the new packet class from TCPPacket
 *         - Implement a default constructor and a constructor taking a
 *           const TCPPacket& argument
 *         - Implement all virtual functions
 *    -# Extend the enum Request in TCPPacket for the new packet type
 *    -# Extend the switch statement in Comm::tcpRecvPacket()
 */
struct TCPPacket : public Serializable {
		//! Types of TCPPackets
		enum Request { UNKNOWN = 0, INFO, CONNECT, PLAYER, WORLD, CHAT, QUIT };
		
		//! Type of the packet
		Request request;
		//! Size of the packet
		Uint32 dataSize;
		
		//! Constructor
		TCPPacket( Request r = UNKNOWN ) : request( r ) { }
		//! Destructor
		virtual ~TCPPacket() { }
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return 0;
		}		
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) { }
		/*! \return The overall size of the packet
		 */
		static Uint32 getStaticSerializeBufferSize() {
			return Serialize<Request>::sizeOf() +
				Serialize<Uint32>::sizeOf();
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			Serialize<Request>::serialize( request, bufferPointer );
			Serialize<Uint32>::serialize( dataSize, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			Serialize<Request>::deserialize( bufferPointer, request );
			Serialize<Uint32>::deserialize( bufferPointer, dataSize );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return getStaticSerializeBufferSize();
		}
};

/*! \brief A specialized packet class for requests for information
 *    about the game running on a server
 */
struct TCPInfoPacket : public TCPPacket {
		//! The game theme
		String theme;
		//! The number of players
		Uint8 numberPlayers;
		
		//! Constructor
		TCPInfoPacket()
			: TCPPacket( INFO ), theme( "dummy" ) { }
		//! Constructor
		TCPInfoPacket( const TCPPacket& packet )
			: TCPPacket( packet ) { }
		//! Destructor
		virtual ~TCPInfoPacket() { }

		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return theme.getSerializeBufferSize() +
				Serialize<Uint8>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			theme.deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, numberPlayers );
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			theme.serialize( bufferPointer );
			Serialize<Uint8>::serialize( numberPlayers, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			theme.deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, numberPlayers );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return TCPPacket::getStaticSerializeBufferSize()
				+ theme.getSerializeBufferSize()
				+ Serialize<Uint8>::sizeOf();
		}
};

/*! \brief A specialized packet class for connection requests
 *    from clients
 */
struct TCPConnectPacket : public TCPPacket {
		//! The local UDP port of the client
		Uint16 port;
		
		//! Constructor
		TCPConnectPacket() : TCPPacket( CONNECT ), port( 0 ) { }
		//! Constructor
		TCPConnectPacket( const TCPPacket& packet )
			: TCPPacket( packet ) { }
		//! Destructor
		virtual ~TCPConnectPacket() { }
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return Serialize<Uint16>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			Serialize<Uint16>::deserialize( bufferPointer, port );
		}
		/*! \return The overall size of the packet
		 */
		static Uint32 getStaticSerializeBufferSize() {
			return TCPPacket::getStaticSerializeBufferSize()
				+ Serialize<Uint16>::sizeOf();
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			Serialize<Uint16>::serialize( port, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			Serialize<Uint16>::deserialize( bufferPointer, port );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return getStaticSerializeBufferSize();
		}
};

/*! \brief A specialized packet class for transmitting a player
 */
struct TCPPlayerPacket : public TCPPacket {
		//! The player to be transmitted
		Player* player;
		//! The frame where the player enters the world
		Uint8 frame;
		
		//! Constructor
		TCPPlayerPacket() : TCPPacket( PLAYER ), player( 0 ), frame( 0 ) { }
		//! Constructor
		TCPPlayerPacket( const TCPPacket& packet )
			: TCPPacket( packet ), player( 0 ), frame( 0 ) { }
		//! Destructor
		virtual ~TCPPlayerPacket() {
			// to delete player or not to delete... that is the question
		}
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return player->getSerializeBufferSize() +
				Serialize<Uint8>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			if ( !player ) {
				player = new Player();
			}
			player->deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			if ( player ) {
				player->serialize( bufferPointer );
			}
			Serialize<Uint8>::serialize( frame, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			if ( !player ) {
				player = new Player();
			}
			player->deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return TCPPacket::getSerializeBufferSize()
				+ player->getSerializeBufferSize()
				+ Serialize<Uint8>::sizeOf();
		}
};

/*! \brief A specialized packet class for transmitting a world
 */
struct TCPWorldPacket : public TCPPacket {
		//! The world to be transmitted
		World* world;
		//! The current frame number for the world
		Uint8 frame;
		//! The player ID for the new player
		Uint8 playerID;
		
		//! Constructor
		TCPWorldPacket() : TCPPacket( WORLD ), world( 0 ) { }
		//! Constructor
		TCPWorldPacket( const TCPPacket& packet )
			: TCPPacket( packet ), world( 0 ) { }
		//! Destructor
		virtual ~TCPWorldPacket() {
			// to delete world or not to delete... that is the question
		}
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return world->getSerializeBufferSize() +
				2 * Serialize<Uint8>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			if ( !world ) {
				world = new World();
			}
			world->deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
			Serialize<Uint8>::deserialize( bufferPointer, playerID );
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			if ( world ) {
				world->serialize( bufferPointer );
			}
			Serialize<Uint8>::serialize( frame, bufferPointer );
			Serialize<Uint8>::serialize( playerID, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			if ( !world ) {
				world = new World();
			}
			world->deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
			Serialize<Uint8>::deserialize( bufferPointer, playerID );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return TCPPacket::getSerializeBufferSize()
				+ world->getSerializeBufferSize()
				+ 2 * Serialize<Uint8>::sizeOf();
		}
};

/*! \brief A specialized packet class for transmitting a chat message
 */
struct TCPChatPacket : public TCPPacket {
		//! The chat message to be transmitted
		String message;
		SDL_Color color;
		
		//! Constructor
		TCPChatPacket() : TCPPacket( CHAT ) { }
		//! Constructor
		TCPChatPacket( const TCPPacket& packet )
			: TCPPacket( packet ) { }
		//! Destructor
		virtual ~TCPChatPacket() { }
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return message.getSerializeBufferSize() +
				3 * Serialize<Uint8>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			message.deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, color.r );
			Serialize<Uint8>::deserialize( bufferPointer, color.g );
			Serialize<Uint8>::deserialize( bufferPointer, color.b );
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			message.serialize( bufferPointer );
			Serialize<Uint8>::serialize( color.r, bufferPointer );
			Serialize<Uint8>::serialize( color.g, bufferPointer );
			Serialize<Uint8>::serialize( color.b, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			message.deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, color.r );
			Serialize<Uint8>::deserialize( bufferPointer, color.g );
			Serialize<Uint8>::deserialize( bufferPointer, color.b );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return TCPPacket::getSerializeBufferSize()
				+ message.getSerializeBufferSize()
				+ 3 * Serialize<Uint8>::sizeOf();
		}
};

struct TCPQuitPacket : public TCPPacket {
		//! The number of the player that has quit
		Uint8 player;
		//! The frame when the player has quit
		Uint8 frame;
		
		//! Constructor
		TCPQuitPacket() : TCPPacket( QUIT ) { }
		//! Constructor
		TCPQuitPacket( const TCPPacket& packet )
			: TCPPacket( packet ) { }
		//! Destructor
		virtual ~TCPQuitPacket() { }
		
		/*! \return Size of the \b data (not counting the headers!) contained in
		 *    the packet
		 */
		virtual Uint32 getDataSize() const {
			return 2 * Serialize<Uint8>::sizeOf();
		}
		/*! Deserialize the data contained in the packet in the buffer starting
		 *  at \em bufferPointer
		 *  \param bufferPointer pointer to the buffer
		 */
		virtual void deserializeData( Uint8*& bufferPointer ) {
			Serialize<Uint8>::deserialize( bufferPointer, player );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
		}
		/*! Serialize the whole packet into the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void serialize( Uint8*& bufferPointer ) const {
			TCPPacket::serialize( bufferPointer );
			Serialize<Uint8>::serialize( player, bufferPointer );
			Serialize<Uint8>::serialize( frame, bufferPointer );
		}
		/*! Deserialize the whole packet from the buffer starting at
		 *  \em bufferPointer
		 */
		virtual void deserialize( Uint8*& bufferPointer ) {
			TCPPacket::deserialize( bufferPointer );
			Serialize<Uint8>::deserialize( bufferPointer, player );
			Serialize<Uint8>::deserialize( bufferPointer, frame );
		}
		/*! \return The size needed for a buffer to contain all data
		 *    for serialization
		 */
		virtual Uint32 getSerializeBufferSize() const {
			return TCPPacket::getSerializeBufferSize()
				+ 2 * Serialize<Uint8>::sizeOf();
		}
};

#endif // _PACKET_HPP_
