/* $Id: player.cpp,v 1.43 2005/05/05 11:52:20 chfreund Exp $ */

#include "player.hpp"

#include <SDL_image.h>

#include "global.hpp"
#include "serialize.hpp"
#include "wopsettings.hpp"
#include "loader.hpp"
#include "avatar.hpp"
#include "weaponshotgun.hpp"
#include "weaponuzi.hpp"
#include "weaponrubgun.hpp"
#include "weapongrenade.hpp"
#include "weaponmine.hpp"
#include "weaponmissile.hpp"
#include "weaponhomingmissile.hpp"
#include "weaponzapper.hpp"
#include "weaponhook.hpp"
#include "weaponhookrel.hpp"
#include "weaponhookcut.hpp"

/**********************************************************/

void Player::updateAvatarData()
{
	DBG(1) if( !CHECK(m_avatar != NULL, "Player::updateAvatarData(): "
	                  "avatar == NULL\n") )  return;

	setPlayerColor( m_playerColor );
}

/**********************************************************/

Player::Player()
      : m_playerColor( 0x00ffffff ),
        m_teamID( 0 ),
        m_avatarType( AVATAR_WORM ),
        m_isLocalPlayer( false ),
        m_avatar( NULL )
{
	// don't forget to increase m_MAX_WEAPONS in player.hpp
	m_weapon[ 0] = new WeaponShotgun();
	m_weapon[ 1] = new WeaponUzi();
	m_weapon[ 2] = new WeaponRubGun();
	m_weapon[ 3] = new WeaponGrenade();
	m_weapon[ 4] = new WeaponMissile();
	m_weapon[ 5] = new WeaponHomingMissile();
	m_weapon[ 6] = new WeaponZapper();
	m_weapon[ 7] = new WeaponMine();
	m_weapon[ 8] = new WeaponHook();
	m_weapon[ 9] = new WeaponHookRel();
	m_weapon[10] = new WeaponHookCut();
	// don't forget to increase m_MAX_WEAPONS in player.hpp

	m_activeWeaponIndex = 0;
}

/**********************************************************/

Player::~Player()
{
	DBG( 1 ) CHECK( m_avatar == NULL, "Player::~Player: m_avatar "
	                "should be detached before deleting the Player "
	                "object; might generate memory hole\n" );
	// cleanup all weapons
	for ( int i = 0; i < m_MAX_WEAPONS; i++ )
		delete m_weapon[i];
}

/**********************************************************/

void Player::update()
{
}

/**********************************************************/

void Player::attachAvatar( Avatar* const avatar )
{
	DBG( 1 ){
		const char fn[] = "Player::attachAvatar:";
		ASSERT( avatar != NULL, "%s passed avatar == NULL\n", fn );
		ASSERT( NULL == m_avatar, "%s attaching avatar to a player, "
		        "that already has one\n", fn );
		ASSERT( NULL == avatar->m_Player, "%s avatar already has an "
		        "attached player\n", fn );
	}

	// plug in bidirectional connection
	m_avatar = avatar;
	avatar->m_Player = this;
	// update the data of the avatar
	updateAvatarData();
}

/**********************************************************/

void Player::setPlayerColor( const Uint32 color )
{
	m_playerColor = color;
	if( m_avatar ) {
		m_avatar->setPlayerColor( m_playerColor );

		// \todo Team colors should be set somewhere else.
		const Uint32 TEAM_COLOR[MAX_NUMBER_OF_PLAYERS] = {
			0x00dd0000, 0x0000dd00, 0x006666ff, 0x00ffff00, 0x00ff00ff, 0x0000ffff,
			0x00990000, 0x00009900, 0x004444aa, 0x00aaaa00, 0x00aa00aa, 0x0000aaaa,
			0x00440000, 0x00004400, 0x00222255, 0x00555500, 0x00550055, 0x00005555,
			0x00ffffff, 0x00aaaaaa
		};

		m_avatar->setTeamColor( TEAM_COLOR[getTeamID()] );
	}
}
		
/**********************************************************/

Avatar* Player::detachAvatar()
{
	LOG( 2 ) INFO( "Player::detachAvatar: detaching avatar\n" );
	DBG( 2 ) {
		ASSERT( m_avatar != NULL,
		        "Player::detachAvatar: m_avatar == NULL\n" );
		ASSERT( m_avatar->m_Player == this,
		        "Player::detachAvatar: avatar did not have me "
		        "as its player\n" );
	}

	// cut the bidirectional connection
	Avatar* temp = m_avatar;
	m_avatar->m_Player = NULL;
	m_avatar = NULL;
	return temp;
}

/**********************************************************/

void Player::serialize( Uint8*& bufferPointer ) const
{
	START_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	// player name
	m_name.serialize( bufferPointer, -1 );
	Serialize<Uint32>::serialize( m_playerColor,       bufferPointer );
	Serialize<Uint8 >::serialize( m_teamID,            bufferPointer );
	Serialize<Uint32>::serialize( m_avatarType,        bufferPointer );
	Serialize<Uint8 >::serialize( m_playerID,          bufferPointer );
	Serialize<Uint8 >::serialize( m_activeWeaponIndex, bufferPointer );

	for ( int i = 0; i < m_MAX_WEAPONS; i++ ) {
		m_weapon[i]->serialize( bufferPointer );
	}

	END_SERIALIZED_SIZE_CHECK( bufferPointer, Player );
}

/**********************************************************/

void Player::deserialize( Uint8*& bufferPointer )
{
	// player name
	m_name.deserialize( bufferPointer );
	Serialize<Uint32>::deserialize( bufferPointer, m_playerColor       );
	Serialize<Uint8 >::deserialize( bufferPointer, m_teamID            );
	Serialize<Uint32>::deserialize( bufferPointer, m_avatarType        );
	Serialize<Uint8 >::deserialize( bufferPointer, m_playerID          );
	Serialize<Uint8 >::deserialize( bufferPointer, m_activeWeaponIndex );
	
	for ( int i = 0; i < m_MAX_WEAPONS; i++ ) {
		m_weapon[i]->deserialize( bufferPointer );
	}
}

/**********************************************************/

Uint32 Player::getSerializeBufferSize() const {
	Uint32 weaponSize = 0;
	for ( int i = 0; i < m_MAX_WEAPONS; i++ ) {
		weaponSize += m_weapon[i]->getSerializeBufferSize();
	}

	return m_name.getSerializeBufferSize()   + // size of the player's name
		Serialize<Uint32>::sizeOf( m_playerColor       ) +
		Serialize<Uint8 >::sizeOf( m_teamID            ) +
		Serialize<Uint32>::sizeOf( m_avatarType        ) +
		Serialize<Uint8 >::sizeOf( m_playerID          ) +
		Serialize<Uint8 >::sizeOf( m_activeWeaponIndex ) +
		weaponSize;
}

/**********************************************************/
